/***********************************************************************

PROJECT:		CS 298 Project

FILE:			edgedetector.cpp

PURPOSE:		Implementation for Boie-Cox edge detector to
				perform edge detection on a grayscale image.

COMPILER:		Visual C++ .NET 2003

PROGRAMMER:		Wallun Chan

START DATE:		9/1/05

REMARKS:		This code is taken from
				http://www.ee.ucl.ac.uk/~icox/#Anchor--Ima-25529.
				The code was originally written in C, but was
				converted into slightly more object-oriented 
				C++ code.

***********************************************************************/

#include "stdafx.h"
#include "edgedetector.h"
#include <math.h>

const int EDGE_X = 252;
const int EDGE_Y = 253;
const int EDGE_45 = 254;
const int EDGE_135 = 255;
const int FWD = 0;
const int BWD = 1;
const double diag_scale = 1.414;

CEdgeDetector::CEdgeDetector(BYTE* array, int xdim, int ydim)
	: an_image(array), nx(xdim), ny(ydim)
/* PURPOSE:		Default constructor to initialize detector 
				member variables.

   RECEIVES:	Image data in byte array, width, and 
				height of image.
*/
{
	gaussian = 0;
	filter = 0;
	idx = 0;
	idy = 0;
	id45 = 0;
	id135 = 0;
	edge_map = 0;
	norm = 0;
	nf = 0;
	nf2 = 0;
	orient_flag = 0;
	lo_threshold = 0;	
	edge_map_hi = 0;
}

CEdgeDetector::~CEdgeDetector()
// Destructor
{
	freeImage();
}

void CEdgeDetector::freeImage()
// Release allocated memory
{
	delete [] gaussian;
	delete [] filter;
	delete [] idx;
	delete [] idy;
	delete [] id45;
	delete [] id135;
	delete [] edge_map_hi;
	delete [] an_image;
	gaussian = 0;
	filter = 0;
	idx = 0;
	idy = 0;
	id45 = 0;
	id135 = 0;
	norm = 0;
	nf = 0;
	nf2 = 0;
	orient_flag = 0;
	lo_threshold = 0;	
	nx = 0;
	ny = 0;
	edge_map = 0;
	edge_map_hi = 0;
	an_image = 0;
}

void CEdgeDetector::setImage(BYTE* array, int xdim, int ydim)
/* PURPOSE:		Initializes detector with image data to be
				edge detected.

   RECEIVES:	Image data in byte array, width, and 
				height of image.
*/
{
	freeImage();
	an_image = array;
	nx = xdim;
	ny = ydim;
}

unsigned char* CEdgeDetector::edge_detect(double sigma, int* threshold)
{
	if (nx == 0 || ny == 0)
		return 0;
	image_find_edges(sigma, threshold);
	image_hysteresis(*threshold / 3);
	return image_thin();
}

unsigned char* CEdgeDetector::image_find_edges(double sigma, int* threshold)
/* PURPOSE:		Generates edge data for image.

   RECEIVES:	Double value for sigma that determines the width
				of the Gaussian filter (increasing this value
				provides more smoothing, which may decrease
				edge detail), and integer value for the threshold,
				which determines how sensitive the detector is
				to changes in intensity (increasing this
				value decreases edge content).

   RETURNS:		Edge data of image.

   REMARKS:		If a zero is input for threshold, then a best
				effort threshold is estimated.
*/
{
	generate_filter(sigma);
	image_convolve(an_image);
	image_detx();
	image_dety();
	image_det45();
	image_det135();
	if (*threshold == 0) 
		*threshold = image_threshold();
	edge_map = image_edges(*threshold);
	return edge_map;
}

void CEdgeDetector::generate_filter(double sigma)
{
	double two_sigma_sq;
	nf = (int)(6 * sigma);
	nf = (nf % 2 > 0 ? nf : nf + 1);
	nf2 = nf / 2;
	filter = new int[nf];
	two_sigma_sq = (double)(2.0 * sigma * sigma);
	norm = 0;
	for (int i = 0; i < nf; i++) 
	{	double x = (double)i - nf2;
		filter[i] = (int)(255.0 * exp(-x * x / two_sigma_sq));
		norm += filter[i];
	}
	norm *= norm;
}

void CEdgeDetector::image_convolve(BYTE* pic)
{
	int ik;
	int ix;
	int iy;
	int nxX2M1;
	int iklim;
	int ikXnx;
	int nfXnx;
	int nyX2M1Xnx;
	int iyMnf2Xnx;
	int iy0;
	int iy0Xnx;
	int it;
	int result;
	int* filtP;
	int* tmpP;
	int* gaussianP;
	int* temp_image;
	BYTE* tpicP;
	BYTE* picP;
	nfXnx = nf * nx;
	nxX2M1 = nx + nx - 1;
	nyX2M1Xnx = (ny + ny - 1) * nx;
	temp_image = new int[nx * ny];
	gaussian = new int[nx * ny];
	tmpP = &temp_image[0];
	picP = pic;
	for (iy = 0; iy < ny; iy++, tmpP += nx, picP += nx) 
	{	for (ix = 0; ix < nf2; ix++) 
		{	filtP = filter;
			result = 0;
			ik = ix - nf2;
			iklim = ik + nf;
			for (; ik < iklim; ik++)
				result += (*filtP++) * picP[abs(ik)];
			*(tmpP + ix) = result;
		}
	}
	tmpP = &temp_image[0];
	tpicP = pic;
	for (iy = 0; iy < ny; iy++, tmpP += nx, tpicP += nx) 
	{	int xlim = nx - nf2 - 1;
		for (ix = nf2; ix < xlim; ix++) 
		{	picP = tpicP + ix - nf2;
			filtP = filter;
			result = 0;
			ik = nf;
			while (ik--)
				result += (*filtP++) * (*picP++);
			*(tmpP + ix) = result;
		}
	}
	tmpP = &temp_image[0];
	picP = pic;
	for (iy = 0; iy < ny; iy++, tmpP += nx, picP += nx) 
	{	for (ix = nx - nf2 - 1; ix < nx; ix++) 
		{	filtP = filter;
			result = 0;
			ik = ix - nf2;
			iklim = ik + nf;
			for (; ik < iklim; ik++) 
			{	if (ik >= nx)
					result += *filtP * picP[nxX2M1 - ik];
				else
					result += *filtP * picP[ik];
				filtP++;
			}
			*(tmpP + ix) = result;
		}
	}
	it = -nf2 * nx;
	for (ix = 0; ix < nx; ix++) 
	{	gaussianP = &gaussian[ix];
		tmpP = &temp_image[ix];
		iy = 0;
		iyMnf2Xnx = it; 
		for (; iy < nf2; iy++, iyMnf2Xnx += nx) 
		{	filtP = filter;
			result = 0;
			ik = iyMnf2Xnx;
			iklim = ik + nfXnx;
			for (; ik < iklim; ik += nx)
				result += (*filtP++) * tmpP[abs (ik)];
			*(gaussianP) = result / norm;
			gaussianP += nx;
		}
	}
	iy0 = nf2;
	iy0Xnx = iy0 * nx;
	for (ix = 0; ix < nx; ix++) 
	{	int* col_tmpP = &temp_image[ix];
		int ylim = ny - nf2 - 1;
		iy = iy0;
		gaussianP = &gaussian[iy0Xnx + ix];
		iyMnf2Xnx = 0;
		for (; iy < ylim; iy++, iyMnf2Xnx += nx) 
		{	filtP = filter;
			tmpP = col_tmpP + iyMnf2Xnx;
			result = 0;
			ik = nf;
			while (ik--) 
			{	result += (*filtP++) * (*tmpP);
				tmpP += nx;
			}
			*(gaussianP) = result / norm;
			gaussianP += nx;
		}
	}
	iy0 = ny - nf2 - 1;
	iy0Xnx = iy0 * nx;
	it = (iy0 - nf2) * nx;
	for (ix = 0; ix < nx; ix++) 
	{	tmpP = &temp_image[ix];
		iy = iy0;
		gaussianP = &gaussian[iy0Xnx + ix];
		iyMnf2Xnx = it;
		for (; iy < ny; iy++, iyMnf2Xnx += nx) 
		{	filtP = filter;
			result = 0;
			ik = iy - nf2;
			iklim = ik + nf;
			ikXnx = iyMnf2Xnx;
			for (; ik < iklim; ik++) 
			{	if (ik >= ny) 
					result += *filtP * tmpP[nyX2M1Xnx - ikXnx];
				else
					result += *filtP * tmpP[ikXnx];
				filtP++;
				ikXnx += nx;
			}
			*(gaussianP) = result / norm;
			gaussianP += nx;
		}
	}
	delete [] temp_image;
}

int CEdgeDetector::image_zc_x(int ix, int iy, int i)
{ 
	return ((image_locx(iy, ix, i) > 0) != 
		(image_locx(iy, ix + 1, i + 1) > 0)) ? EDGE_X : 0;
}

int CEdgeDetector::image_zc_y(int ix, int iy, int i)
{
	return ((image_locy(iy + 1, ix, i + nx) > 0) != 
		(image_locy(iy, ix, i) > 0)) ? EDGE_Y : 0;
}

int CEdgeDetector::image_zc_135(int ix, int iy, int i)
{
	return ((image_loc135(iy + 1, ix + 1, i + nx + 1) > 0) != 
		(image_loc135(iy, ix, i) > 0)) ? EDGE_135 : 0;
}

int CEdgeDetector::image_zc_45(int ix, int iy, int i)
{
	return ((image_loc45(iy + 1,ix - 1, i + nx - 1) > 0) != 
		(image_loc45(iy, ix, i) > 0)) ? EDGE_45 : 0;
}

unsigned char* CEdgeDetector::image_edges(int threshold)
{
	int ix;
	int iy;
	int orient_flag;
	int index;
	int* px = idx;
	int* py = idy;
	int* p45 = id45;
	int* p135 = id135;
	BYTE* mp;
	BYTE* mapp;
	mp = new BYTE[nx * ny];
	mapp = mp;
	for (iy = 0; iy < ny; iy++) 
	{	for (ix = 0; ix < nx; ix++) 
		{	int sx;
			int sy;
			int s45;
			int s135;
			int big;
			sx = abs(*px);
			px++;
			sy = abs(*py);
			py++;
			s45 = (int)((double)(abs(*p45)) / diag_scale);
			p45++;
			s135 = (int)((double)(abs(*p135)) / diag_scale);
			p135++;
			big = sx;
			orient_flag = EDGE_X;
			if (sy > big) 
			{	big = sy;
				orient_flag = EDGE_Y;
			}
			if (s45 > big) 
			{	big = s45;
				orient_flag = EDGE_45;
			}
			if (s135 > big) 
			{	big = s135;
				orient_flag = EDGE_135;
			}
			orient_flag = orient_flag;
			if (big > threshold) 
			{	index = (int)(mapp - mp);
				switch (orient_flag) 
				{
					case EDGE_X:
						*mapp++ = image_zc_x(ix, iy, index);
						break;
					case EDGE_Y:
						*mapp++ = image_zc_y(ix, iy, index);
						break;
					case EDGE_45:
						*mapp++ = image_zc_45(ix, iy, index);
						break;
					case EDGE_135:
						*mapp++ = image_zc_135(ix, iy, index);
						break;
					default:
						throw "CEdgeDetector::image_edges() - error in case statements.";
				}
			}
			else 
				*mapp++ = 0;
		}
	}
	return mp;
}

int CEdgeDetector::image_edge_map_lo(int ix, int iy, int index)
{
	int big;
	int sx;
	int sy;
	int s45;
	int s135;
	sx = abs(idx[index]);
	sy = abs(idy[index]);
	s45 = (int)((double)(abs(id45[index])) / diag_scale);
	s135 = (int)((double)(abs(id135[index])) / diag_scale);
	big = sx;
	orient_flag = EDGE_X;
	if (sy > big) 
	{	big = sy;
		orient_flag = EDGE_Y;
	}
	if (s45 > big) 
	{	big = s45;
		orient_flag = EDGE_45;
	}
	if (s135 > big) 
	{	big = s135;
		orient_flag = EDGE_135;
	}
	if (big > lo_threshold) 
	{	switch (orient_flag) 
		{
			case EDGE_X:
				return image_zc_x(ix, iy, index);
			case EDGE_Y:
				return image_zc_y(ix, iy, index);
			case EDGE_45:
				return image_zc_45(ix, iy, index);
			case EDGE_135:
				return image_zc_135(ix, iy, index);
			default:
				throw "CEdgeDetector::image_edge_map_lo() - error in case statements.";
		}
	}
	else
		return 0;
}

void CEdgeDetector::image_detx()
{
	int* dP;
	int* gP;
	int ix;
	int iy;
	int nxM1;
	nxM1 = nx - 1;
	idx = new int[nx * ny];
	dP = idx;
	gP = gaussian;
	for (iy = 0; iy < ny; iy++) 
	{	*dP = *(gP + 1) - *(gP);
		dP += nx;
		gP += nx;
	}
	for (iy = 0; iy < ny; iy++) 
	{	dP = idx + iy * nx + 1;
		gP = gaussian + iy * nx + 2;
		for (ix = 1; ix < nxM1; ix++) 
		{	*dP = *gP - *(gP - 2);
			dP++;
			gP++;
		}
	}
	dP = idx + nxM1;
	gP = gaussian + nxM1;
	for (iy = 0; iy < ny; iy++) 
	{	*dP = *gP - *(gP - 1);
		dP += nx;
		gP += nx;
	}
}

void CEdgeDetector::image_dety()
{
	int* dP;
	int* gP;
	int ix;
	int iy;
	int nyM1;
	int nxPnx;
	nxPnx = nx + nx;
	nyM1 = ny - 1;
	idy = new int[nx * ny];
	dP = idy;
	gP = gaussian;
	for (ix = 0; ix < nx; ix++) 
		dP[ix] = gP[ix + nx] - gP[ix];
	for (ix = 0; ix < nx; ix++) 
	{	dP = idy + ix;
		gP = gaussian + ix;
		for (iy = 1; iy < nyM1; iy++) 
		{	dP += nx;
			*dP = *(gP + nxPnx) - *gP;
			gP += nx;
		}
	}
	dP = idy + (ny - 1) * nx;
	gP = gaussian + (ny - 1) * nx;
	for (ix = 0; ix < nx; ix++) 
		dP[ix] = gP[ix] - gP[ix - nx];
}

void CEdgeDetector::image_det45()
{
	int* dP;
	int* gP;
	int i;
	int ix;
	int iy;
	int index;
	int nxP1;
	int nxM1;
	int nyM1;
	nxM1 = nx - 1;
	nxP1 = nx + 1;
	nyM1 = ny - 1;
	id45 = new int[nx * ny];
	dP = id45 + nx - 2;
	gP = gaussian + nx - 2;
	for (ix = nx - 2; ix > 0; ix--) 
	{	*dP = *(gP + nxM1) - *(gP + 1);
		dP--;
		gP--;
	}
	dP = id45;
	gP = gaussian;
	for (i = nx - 2; i > 0; i--) 
	{	ix = i;
		iy = 1;
		index = nx + ix;
		while (ix > -1 && iy < nyM1) 
		{	dP[index] = gP[index + nxM1] - gP[index - nxM1];
			index += nxM1;
			ix--;
			iy++;
		}
	}
	for (i = 1; i < ny; i++) 
	{	ix = nx - 2;
		iy = i;
		index = iy * nx + ix;
		while (ix > 0 && iy < nyM1) 
		{	dP[index] = gP[index + nxM1] - gP[index - nxM1];
			index += nxM1;
			ix--;
			iy++;
		}
	}
	dP = id45 + nx + nx;
	gP = gaussian + nx + nx;
	for (iy = 1; iy < nyM1; iy++) 
		*dP = *(gP + nx) - *(gP - nxM1);
	dP = id45 + nx + nx - 1;
	gP = gaussian + nx + nx - 1;
	for (iy = 1; iy < nyM1; iy++) 
	{	*dP = *(gP + nxM1) - *(gP - nx);
		dP += nx;
		gP += nx;
	}
	dP = id45 + (ny - 1) * nx;
	gP = gaussian + (ny - 1) * nx;
	for (ix = 1; ix < nxM1; ix++) 
	{	*dP = *(gP - 1) - *(gP - nxM1);
		dP++;
		gP++;
	}
	dP = id45;
	gP = gaussian;
	*dP = *(gP + nx) - *(gP + 1);
	dP = id45 + (ny - 1) * nx;
	gP = gaussian + (ny - 1) * nx;
	*dP = *(gP) - *(gP - nxM1);
	dP[nxM1] = *(gP + nxM1 - 1) - *(gP - 1);
	dP = id45 + nxM1;
	gP = gaussian + nxM1;
	*dP = *(gP + nxM1) - *(gP);
}

void CEdgeDetector::image_det135()
{
	int* dP;
	int* gP;
	int i;
	int ix;
	int iy;
	int index;
	int nxP1;
	int nxM1;
	int nyM1;
	int nxM2;
	nxM1 = nx - 1;
	nxP1 = nx + 1;
	nxM2 = nx - 2;
	nyM1 = ny - 1;
	id135 = new int[nx * ny];
	dP = id135 + 1;
	gP = gaussian + 1;
	for (ix = 1; ix < nxM1; ix++) 
	{	*dP = *(gP + nxP1) - *(gP - 1);
		dP++;
		gP++;
	}
	dP = id135;
	gP = gaussian;
	for (i = 1; i < nxM1; i++) 
	{	ix = i;
		iy = 1;
		index = nx + ix;
		while (ix < nxM1 && iy < nyM1) 
		{	dP[index] = gP[index + nxP1] - gP[index - nxP1];
			index += nxP1;
			ix++;
			iy++;
		}
	}
	for (i = 1; i < ny; i++) 
	{	ix = 1;
		iy = i;
		index = iy * nx + ix;
		while (ix < nxM1 && iy < nyM1) 
		{	dP[index] = gP[index + nxP1] - gP[index - nxP1];
			index += nxP1;
			ix++;
			iy++;
		}
	}
	dP = id135 + nx;
	gP = gaussian + nx;
	for (iy = 1; iy < nyM1; iy++) 
	{	*dP = *(gP + nxP1) - *(gP - nx);
		dP += nx;
		gP += nx;
	}
	dP = id135 + nx + nx - 1;
	gP = gaussian + nx + nx - 1;
	for (iy = 1; iy < nyM1; iy++) 
	{	*dP = *(gP + nx) - *(gP - nxP1);
		dP += nx;
		gP += nx;
	}
	dP = id135 + (ny - 1) * nx + 1;
	gP = gaussian + (ny - 1) * nx + 1;
	for (ix = 1; ix < nxM1; ix++) 
	{	*dP = *(gP + 1) - *(gP - nxP1);
		dP++;
		gP++;
	}
	dP = id45;
	gP = gaussian;
	*dP = *(gP + nxP1) - *(gP);
	dP = id45 + (ny - 1) * nx;
	gP = gaussian + (ny - 1) * nx;
	*dP = *(gP + 1) - *(gP - nx);
	dP[nxM1] = *(gP + nxM1) - *(gP - 2);
	dP = id45 + nxM1;
	gP = gaussian + nxM1;
	*dP = *(gP + nx) - *(gP - 1);
}

unsigned char* CEdgeDetector::image_hysteresis(int lo_thld)
{
	int iy;
	int ix;
	BYTE* lmapP;
	BYTE* lmap_hiP;
	BYTE* mapP;
	lo_threshold = lo_thld;
	edge_map_hi = edge_map;
	edge_map = new BYTE[nx * ny];
	lmap_hiP = &edge_map_hi[0];
	lmapP = &edge_map[0];
	mapP = lmapP;
	for (ix = nx * ny; --ix >= 0;)
		*mapP++ = 0;
	for (iy = 0; iy < ny; iy++, lmap_hiP += nx, lmapP += nx) 
	{	mapP = lmapP;
		for (ix = 0; ix < nx; ix++, mapP++) 
		{	if (lmap_hiP[ix] != 0) 
			{	*mapP = lmap_hiP[ix];
				image_neighbor_fwd(iy, ix, mapP);
				image_neighbor_bwd(iy, ix, mapP);
			}
		}
	}
	return edge_map;
}

unsigned char* CEdgeDetector::image_thin()
{
	int ix;
	int iy;	
	int nxM1;
	int nxP1;
	int nlinks;
	int npieces;
	int nyM1;
	int b01;
	int b12;
	int b21;
	int b10;
	int p1;
	int p2;
	int p3;
	int p4;
	int b00;
	int b02;
	int b20;
	int b22;	
	BYTE* emP;
	BYTE* emPiy;
	nxM1 = nx - 1;
	nxP1 = nx + 1;
	nyM1 = ny - 1;
	emPiy = &edge_map[0];
	for (iy = 1; iy < nyM1; iy++) 
	{	emPiy += nx;
		for (ix = 1; ix < nxM1; ix++) 
		{	emP = emPiy + ix;
			b01 = *(emP - nx) > 0;
			b12 = *(emP + 1) > 0;
			b21 = *(emP + nx) > 0;
			b10 = *(emP - 1) > 0;
			if ((b01 + b12 + b21 + b10) > 1) 
			{	b00 = *(emP - nxP1) > 0;
				b02 = *(emP - nxM1) > 0;
				b20 = *(emP + nxM1) > 0;
				b22 = *(emP + nxP1) > 0;
				p1 = b00 | b01;
				p2 = b02 | b12;
				p3 = b22 | b21;
				p4 = b20 | b10;
				nlinks = b01 & p2;
				nlinks += b12 & p3;
				nlinks += b21 & p4;
				nlinks += b10 & p1;
				npieces = p1 + p2 + p3 + p4;
				if ((npieces - nlinks) < 2)
					*emP = 0;
			}
		}
	}
	return edge_map;
}

void CEdgeDetector::image_neighbor_fwd(int iy, int ix, BYTE* mapP)
{
	switch (*mapP) 
	{
		case 5:
			return;
		case EDGE_Y:               
			if ((!Fcontour(iy, ix + 1) && Fcontour(iy, ix + 2)) ||
				(!(Fcontour(iy - 1, ix + 1) || Fcontour(iy + 1, ix + 1)) &&
				(Fcontour(iy - 1, ix + 2) || Fcontour(iy + 1, ix + 2))))
			{
				if (ix + 1 < nx)
					*(mapP + 1) = 5;
			}
			break;
		case EDGE_135:            
			if ((!Fcontour(iy - 1, ix + 1) && Fcontour(iy - 2, ix + 2)) ||
				(!(Fcontour(iy, ix + 1) || Fcontour(iy - 1, ix)) &&
				(Fcontour(iy - 2, ix + 1) || Fcontour(iy - 1, ix + 2))))
			{
				if ((iy - 1 >= 0) && (ix + 1 < nx))
					*(mapP - nx + 1) = 5;
			}
			break;
		case EDGE_X:               
			if ((!Fcontour(iy - 1, ix) && Fcontour(iy - 2, ix)) ||
				(!(Fcontour(iy - 1, ix - 1) || Fcontour(iy - 1, ix + 1)) &&
				(Fcontour(iy - 2, ix - 1) || Fcontour(iy - 2, ix + 1))))
			{
				if (iy - 1 >= 0)
					*(mapP - nx) = 5;
			}
			break;
		case EDGE_45:             
			if ((!Fcontour(iy + 1, ix + 1) && Fcontour(iy + 2, ix + 2)) ||
				(!(Fcontour(iy, ix + 1) || Fcontour(iy + 1, ix)) &&
				(Fcontour(iy + 1, ix + 2) || Fcontour(iy + 2, ix + 1))))
			{
				if ((iy + 1 < ny) && (ix + 1 < nx))
					*(mapP + nx + 1) = 5;
			}
			break;
		default:
			throw "CEdgeDetector::image_neighbor_fwd() - error in case statements.";
	}
}

void CEdgeDetector::image_neighbor_bwd(int iy, int ix, BYTE* mapP)
{
	switch (*mapP) 
	{
		case 5:
			return;
		case EDGE_Y:              
			if ((!Bcontour(iy, ix - 1) && Bcontour(iy, ix - 2)) ||
				(!(Bcontour(iy - 1, ix - 1) || Bcontour(iy + 1, ix - 1)) &&
				(Bcontour(iy - 1, ix - 2) || Bcontour(iy + 1, ix - 2))))
			{
				if (ix - 1 >= 0)
					*(mapP - 1) = 5;
			}
			break;
		case EDGE_135:               
			if ((!Bcontour(iy + 1, ix - 1) && Bcontour(iy + 2, ix - 2)) ||
				(!(Bcontour(iy, ix - 1) || Bcontour(iy + 1, ix)) &&
				(Bcontour(iy + 1, ix - 2) || Bcontour(iy + 2, ix - 1))))
			{
				if ((iy + 1 < ny) && (ix - 1 >= 0))
					*(mapP + nx - 1) = 5;
			}
			break;
		case EDGE_X:               
			if ((!Bcontour(iy + 1, ix) && Bcontour(iy + 2, ix)) ||
				(!(Bcontour(iy + 1, ix - 1) || Bcontour(iy + 1, ix + 1)) &&
				(Bcontour(iy + 2, ix - 1) || Bcontour(iy + 2, ix + 1))))
			{
				if (iy + 1 < ny)
					*(mapP + nx) = 5;
			}
			break;
		case EDGE_45:               
			if ((!Bcontour(iy - 1, ix - 1) && Bcontour(iy - 2, ix - 2)) ||
				(!(Bcontour(iy, ix - 1) || Bcontour(iy - 1, ix)) &&
				(Bcontour(iy - 1, ix - 2) || Bcontour(iy - 2, ix - 1))))
			{
				if ((iy - 1 >= 0) && (ix - 1 >= 0))
					*(mapP - nx - 1) = 5;
			}
			break;
		default:
			throw "CEdgeDetector::image_neighbor_bwd() - error in case statements.";
	}
}

int CEdgeDetector::Fcontour(int iy, int ix)
{
	int edge_pt;
	int index;
	BYTE* mapP;
	index = iy * nx + ix;
	if (index < 0 || index >= nx*ny)
		return 0;
	mapP = &edge_map[index];
	if (*mapP > 0)
		return 1;
	else if ((edge_pt = image_edge_map_lo(ix, iy, index)) != 0) 
	{	*mapP = edge_pt;
		image_neighbor_fwd(iy, ix, mapP);
		return 1;
	}
	else
		return 0;
}

int CEdgeDetector::Bcontour(int iy, int ix)
{
	BYTE* mapP;
	int edge_pt;
	int index;
	index = iy * nx + ix;
	if (index < 0 || index >= nx * ny)
		return 0;
	mapP = &edge_map[index];
	if (*mapP > 0)
		return 1;
	else if ((edge_pt = image_edge_map_lo (ix, iy, index)) != 0) 
	{	*mapP = edge_pt;
		image_neighbor_bwd (iy, ix, mapP);
		return 1;
	}
	else
		return 0;
}

int CEdgeDetector::image_locx(int iy, int ix, int index)
{
	int* imP;
	if (ix < 2 || ix > nx - 2 || iy < 0 || iy > ny - 1)
		return 0;
	imP = &idx[index];
	return *imP - *(imP - 1);
}

int CEdgeDetector::image_locy(int iy, int ix, int index)
{
	int* imP;
	if (ix < 0 || ix > nx - 1 || iy < 2 || iy > ny - 2)
		return 0;
	imP = &idy[index];
	return *imP - *(imP - nx);
}

int CEdgeDetector::image_loc45(int iy, int ix, int index)
{
	int* imP;
	if (ix < 2 || ix > nx - 2 || iy < 2 || iy > ny - 2)
		return 0;
	imP = &id45[index];
	return *imP - *(imP - nx + 1);
}

int CEdgeDetector::image_loc135(int iy, int ix, int index)
{
	int* imP;
	if (ix < 2 || ix > nx - 2 || iy < 2 || iy > ny - 2)
		return 0;
	imP = &id135[index];
	return *imP - *(imP - nx - 1);
}

int CEdgeDetector::image_threshold()
{
	int ix;
	int iy;
	int* detx;
	int tmp;
	double mad = 0.0;
	detx = idx;
	for (iy = 0; iy < ny; iy++) 
	{	for (ix = 0; ix < nx; ix++) 
		{	tmp = *detx++;
			mad += abs(tmp);
		}
	}
	return (int)((double)(3 * mad / nx / ny) / 0.8);
}