/********************************************************************************

PROJECT:		CS 298 Project

FILE:			featurelocate.cpp

PURPOSE:		Finds boundary boxes around face features.

COMPILER:		Visual C++ .NET 2003

PROGRAMMER:		Wallun Chan

START DATE:		9/1/05

REMARKS:		Invokes face detection library taken from
				http://vasc.ri.cmu.edu/NNFaceDetector.

********************************************************************************/

#include "stdafx.h"
#include "featurelocate.h"
#include "edgedetector.h"
#include "constants.h"

CFeatureLocate::CFeatureLocate() : imageMatrix(0), width(0), height(0)
{
}

CFeatureLocate::~CFeatureLocate()
{
}

void CFeatureLocate::setImage(CImageMatrix* image)
{
	imageMatrix = image;
	if (imageMatrix == 0 || imageMatrix->isNull())
	{	width = 0;
		height = 0;
	}	
	else 
	{	width = imageMatrix->getWidth();
		height = imageMatrix->getHeight();
	}
}

BOOL CFeatureLocate::setData(CData* data, double sigma, int* threshold)
{	
	// Get eye and face dimensions by invoking neural network
	FaceEyeLocation* locs = findFace();

	if (locs == 0)
		return FALSE;
	data->width = width;
	data->height = height;

	// Calculate in-plane rotation angle to straighten out face
	int xDif = abs(locs->leftx - locs->rightx);
	int yDif = abs(locs->lefty - locs->righty);
	double angle = atan((double)yDif / xDif);
	data->param.angle = locs->righty > locs->lefty ? -angle : angle;

	// Rotate eye midpoint to calculated rotation angle
	CPoint lPt = rotatePoint(locs->leftx, locs->lefty, -angle, width / 2, height / 2);
	CPoint rPt = rotatePoint(locs->rightx, locs->righty, -angle, width / 2, height / 2);

	int distBtwnEyes = abs(lPt.x - rPt.x);

	// Find locations of eyes, nose, and mouth
	int noseX = (lPt.x + rPt.x) / 2;
	int noseY = (lPt.y + rPt.y + distBtwnEyes) / 2;
	int mouthX = (lPt.x + rPt.x) / 2;
	int mouthY = (lPt.y + rPt.y) / 2 + distBtwnEyes;

	// Put rough boxes around located features
	int radius = distBtwnEyes / 4;
	int eyeMarginX = (int)(EYE_MARGIN_X * radius);
	int eyeMarginY = (int)(EYE_MARGIN_Y * radius);
	CRect regions[FEATURE_NUM - 1];
	regions[0].SetRect(lPt.x - eyeMarginX, lPt.y - eyeMarginY, 
		lPt.x + eyeMarginX, lPt.y + eyeMarginY);
	regions[1].SetRect(rPt.x - eyeMarginX, rPt.y - eyeMarginY, 
		rPt.x + eyeMarginX, rPt.y + eyeMarginY);
	int noseMarginX = (int)(NOSE_MARGIN_X * radius);
	regions[2].SetRect(noseX - noseMarginX, (lPt.y + rPt.y + radius) / 2, 
		noseX + noseMarginX, noseY + radius);
	regions[3].SetRect(mouthX - radius * 2, mouthY - radius, 
		mouthX + radius * 2, mouthY + radius);

	CImageMatrix rotated;
	imageMatrix->rotateImage(rotated, data->param.angle);
	for (int i = FEATURE_NUM - 1; i >= 0; i--)
	{	if (i == 2)
			data->param.rect[i] = regions[i];
		else if (i == FEATURE_NUM - 1)
		{	int left = locs->rightx - distBtwnEyes / 2;
			if (left < 0) 
				left = 0;
			int top = (locs->righty + locs->lefty - distBtwnEyes) / 2;
			if (top < 0) 
				top = 0;
			int right = locs->leftx + distBtwnEyes / 2;
			if (right > width) 
				right = width - 1;
			int bottom = locs->lefty + (int)(NOSE_MARGIN_X * distBtwnEyes);
			if (bottom > height) 
				bottom = height - 1;
			data->param.rect[i].SetRect(left, top, right, bottom);
		}
		else
			data->param.rect[i] = centerBox(rotated, regions[i], sigma, threshold);
	}
	int num = FEATURE_NUM - 1;
	boundRegion(data->param.rect, data->param.rect[num], num);
	for (int j = 0; j < FEATURE_NUM; j++)
		data->param.select[j] = TRUE;

	free(locs);
	return TRUE;
}

void CFeatureLocate::boundRegion(CRect* rects, const CRect& bound, int num)
{
	for (int i = 0; i < num; i++)
	{	rects[i].left = (rects[i].left >= bound.left && 
			rects[i].left <= bound.right) ? rects[i].left : bound.left;
		rects[i].right = (rects[i].right >= bound.left && 
			rects[i].right <= bound.right) ? rects[i].right : bound.right;
		rects[i].top = (rects[i].top >= bound.top && 
			rects[i].top <= bound.bottom) ? rects[i].top : bound.top;
		rects[i].bottom = (rects[i].bottom >= bound.top && 
			rects[i].bottom <= bound.bottom) ? rects[i].bottom : bound.bottom;
	}
}

CRect CFeatureLocate::centerBox(const CImageMatrix& image, 
const CRect& box, double sigma, int* threshold)
{
	width = box.Width() > 0 ? box.Width() + 1 : 0;
	height = box.Height() > 0 ? box.Height() + 1 : 0;
	CEdgeDetector detector(image.getGrayBuffer(box), width, height);
	BYTE* edgeMap = detector.edge_detect(sigma, threshold);
	if (edgeMap == 0)
		return box;
	int xSum = 0;
	int ySum = 0;
	int pixCount = 0;
	int i, j;
	for (i = 0; i < height; i++)
	{	for (j = 0; j < width; j++)
			if (edgeMap[i * width + j] > 0)
			{	ySum += i;
				xSum += j;
				pixCount++;
			}
	}
	int xMean = pixCount == 0 ? width / 2 : xSum / pixCount;
	int yMean = pixCount == 0 ? height / 2 : ySum / pixCount;
	int xDif = xMean - width / 2;
	int yDif = yMean - height / 2;
	CRect rect(box.left + xDif, box.top + yDif, 
		box.right + xDif, box.bottom + yDif);
	detector.setImage(image.getGrayBuffer(rect), width, height);
	edgeMap = detector.edge_detect(sigma, threshold);
	int minX = width;
	int maxX = 0;
	int minY = height;
	int maxY = 0;
	for (i = 0; i < height; i++)
	{	for (j = 0; j < width; j++)
			if (edgeMap[i * width + j] > 0)
			{	if (i < minY)
					minY = i;
				if (i > maxY)
					maxY = i;
				if (j < minX)
					minX = j;
				if (j > maxX)
					maxX = j;
			}
	}
	if (minX > maxX || minY > maxY)
		return CRect(rect.left, rect.top, rect.right, rect.bottom);
	else
		return CRect(minX + rect.left, minY + rect.top, 
			maxX + rect.left, maxY + rect.top);
}

FaceEyeLocation* CFeatureLocate::findFace()
{
	if (imageMatrix == 0)
		return 0;
	FaceEyeLocation* locs = 0;
	BYTE* buffer = imageMatrix->getGrayBuffer();
	int num = Track_FindAllFaces(width, height, buffer, &locs);
	delete [] buffer;
	if (locs == 0)
		return 0;
	if ((num == 1) && (locs->left == 1) && (locs->right == 1) &&
		(locs->leftx >= 0) && (locs->leftx < width) && 
		(locs->rightx >= 0) && (locs->rightx < width) &&
		(locs->lefty >= 0) && (locs->lefty < height) &&
		(locs->righty >= 0) && (locs->righty < height) &&
		(locs->x1 >= 0) && (locs->x1 < width) &&
		(locs->y1 >= 0) && (locs->y1 < height) &&
		(locs->x2 >= 0) && (locs->x2 < width) &&
		(locs->y2 >= 0) && (locs->y2 < height))
		return locs;
	free(locs);
	return 0;
}

CPoint CFeatureLocate::rotatePoint(int x, int y, double angle, int rotX, int rotY)
{
	int xp = abs(x - rotX);
	int yp = abs(y - rotY);
	xp = x < rotX ? -xp : xp;
	yp = y > rotY ? -yp : yp;
	int tmpX = (int)(cos(angle) * xp + sin(angle) * yp);
	int tmpY = (int)(cos(angle) * yp - sin(angle) * xp);
	xp = tmpX + rotX;
	yp = rotY - tmpY;
	return CPoint(xp, yp);
}