/********************************************************************************

PROJECT:		CS 298 Project

FILE:			imagematrix.cpp

PURPOSE:		Implements image scaling, rotation, filtering, mirroring, etc.

COMPILER:		Visual C++ .NET 2003

PROGRAMMER:		Wallun Chan

START DATE:		9/1/05

REMARKS:		Code for image rotation, scaling, and filtering are taken
				from Practical Algorithms for Image Analysis: Description, 
				Examples, and Code.  Michael Seul, Lawrence O'Gorman, 
				and Michael J. Sammon. Cambridge University Press, 2000.

********************************************************************************/

#include "stdafx.h"
#include "imagematrix.h"
#include "edgedetector.h"
#include "polygontester.h"
#include "constants.h"

const double GAUSS_CUTOFF = 0.832555;
const double NORM = 100.0;

CConvexHull CImageMatrix::convexHull;

CImageMatrix::CImageMatrix()
{
}

CImageMatrix::~CImageMatrix()
{
	image.Destroy();
}

BOOL CImageMatrix::createRGBImage(int width, int height, COLORREF color)
{
	if (!image.IsNull())
		image.Destroy();
	BOOL flag = image.Create(width, height, BITS_PER_PIXEL_24);
	if (flag && color != BLACK)
		colorRegion(CRect(0, 0, width, height), color);
	return flag;
}

void CImageMatrix::destroy()
{
	image.Destroy();
}

BOOL CImageMatrix::loadImage(const CString& path)
{
	BOOL succeeded = FALSE;
	if (!image.IsNull())
		image.Destroy();
	HRESULT hr = image.Load(path);
	if (SUCCEEDED(hr) && image.IsDIBSection() && 
		((image.GetBPP() == BITS_PER_PIXEL_8 && 
		image.IsIndexed()) || image.GetBPP() == BITS_PER_PIXEL_24))
		succeeded = TRUE;
	return succeeded;
}

BOOL CImageMatrix::setImage(const CImageMatrix& im)
{
	if (im.image.IsNull())
		return FALSE;
	if (!image.IsNull())
		image.Destroy();
	int width = im.image.GetWidth();
	int height = im.image.GetHeight();
	return setImage(im, 0, 0, 
		width - 1, height - 1, width, height);
}

BOOL CImageMatrix::setImage(const CImageMatrix& im, const CRect& rect)
{
	if (im.image.IsNull())
		return FALSE;
	if (!image.IsNull())
		image.Destroy();
	int width = rect.Width() + 1;
	int height = rect.Height() + 1;
	return setImage(im, rect.left, rect.top, 
		rect.right, rect.bottom, width, height);
}

BOOL CImageMatrix::setImage(const CImageMatrix& copy, 
int left, int top, int right, int bottom, 
int width, int height)
{
	image.Create(width, height, BITS_PER_PIXEL_24);
	if (copy.image.IsIndexed() && copy.image.GetBPP() == BITS_PER_PIXEL_8)
	{	int maxColors = copy.image.GetMaxColorTableEntries();
		RGBQUAD* rgbQuad = new RGBQUAD[maxColors];
		copy.image.GetColorTable(0, maxColors, rgbQuad);
		for (int i = top; i <= bottom; i++)
		{	for (int j = left; j <= right; j++)
			{	BYTE* inAddr = (BYTE*)image.GetPixelAddress(j - left, i - top);
				BYTE copyAddr = *((BYTE*)copy.image.GetPixelAddress(j, i));
				*(inAddr + 2) = rgbQuad[copyAddr].rgbRed;
				*(inAddr + 1) = rgbQuad[copyAddr].rgbGreen;
				*inAddr = rgbQuad[copyAddr].rgbBlue;
			}
		}
		delete [] rgbQuad;
	}
	else if (copy.image.GetBPP() == BITS_PER_PIXEL_24)
	{	for (int i = top; i <= bottom; i++)
		{	for (int j = left; j <= right; j++)
			{	BYTE* inAddr = (BYTE*)image.GetPixelAddress(j - left, i - top);
				BYTE* copyAddr = (BYTE*)copy.image.GetPixelAddress(j, i);
				*inAddr = *copyAddr;
				*(inAddr + 1) = *(copyAddr + 1);
				*(inAddr + 2) = *(copyAddr + 2);
			}
		}
	}
	return TRUE;
}

BOOL CImageMatrix::filterImage(CImageMatrix& im, int fltrSize, double cutOff)
{
	if (image.IsNull())
		return FALSE;
	int width = image.GetWidth();
	int height = image.GetHeight();
	BYTE* rBytes = getColorBuffer(Red);
	BYTE* gBytes = getColorBuffer(Green);
	BYTE* bBytes = getColorBuffer(Blue);
	lowPassFilter(rBytes, width, height, fltrSize, cutOff);
	lowPassFilter(gBytes, width, height, fltrSize, cutOff);
	lowPassFilter(bBytes, width, height, fltrSize, cutOff);
	im.createRGBImage(width, height);
	for (int i = 0; i < height; i++)
	{	for (int j = 0; j < width; j++)
		{	BYTE* addr = (BYTE*)im.image.GetPixelAddress(j, i);
			*addr = rBytes[i * width + j];
			*(addr + 1) = gBytes[i * width + j];
			*(addr + 2) = bBytes[i * width + j];
		}
	}
	delete [] rBytes;
	delete [] gBytes;
	delete [] bBytes;
	return TRUE;
}

BOOL CImageMatrix::isolateFeature(CImageMatrix& im, 
double sigma, int* threshold)
{
	if (image.IsNull())
		return FALSE;
	int width = image.GetWidth();
	int height = image.GetHeight();
	if (!im.createRGBImage(width, height))
		return FALSE;
	CEdgeDetector detector(getGrayBuffer(), width, height);
	BYTE* edgeMap = detector.edge_detect(sigma, threshold);
	if (edgeMap == 0)
		return FALSE;
	VecIntPr pts;
	int i, j;
	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
			if (edgeMap[i * width + j] > 0)
				pts.push_back(Pr(j, i));
	VecIntPr hull;
	convexHull.setHull(pts, hull);
	CPolygonTester polygonTester(hull);
	for (i = 0; i < height; i++)
	{	for (j = 0; j < width; j++)
		{	BYTE* addr1 = (BYTE*)image.GetPixelAddress(j, i);
			BYTE* addr2 = (BYTE*)im.image.GetPixelAddress(j, i);
			if (polygonTester.isPointInPolygon(j, i))
			{	*addr2 = *addr1;
				*(addr2 + 1) = *(addr1 + 1);
				*(addr2 + 2) = *(addr1 + 2);
			}
			else
				*(addr2 + 2) = *(addr2 + 1) = *addr2 = 0;
		}
	}
	return TRUE;
}

void CImageMatrix::overlay(const CPoint& pt, 
const CImageMatrix& im, BOOL blendFlag)
{
	if (im.isNull())
		return;
	int left = __max(0, pt.x);
	int right = __min(pt.x + im.image.GetWidth(), image.GetWidth());
	int top = __max(0, pt.y);
	int bottom = __min(pt.y + im.image.GetHeight(), image.GetHeight());
	for (int i = top; i < bottom; i++)
	{	for (int j = left; j < right; j++)
		{	BYTE* wAddr = (BYTE*)image.GetPixelAddress(j, i);
			BYTE* rAddr = (BYTE*)im.image.GetPixelAddress(j - left, i - top);
			BYTE blue = *rAddr;
			BYTE green = *(rAddr + 1);
			BYTE red = *(rAddr + 2);
			if (!blendFlag || (blendFlag && red + green + blue != 0))
			{	*wAddr = blue;
				*(wAddr + 1) = green;
				*(wAddr + 2) = red;
			}
		}
	}
}

void CImageMatrix::colorRegion(const CRect& rect, COLORREF color)
{
	if (image.IsNull())
		return;
	int left = __max(0, rect.left);
	int right = __min(rect.right + 1, image.GetWidth());
	int top = __max(0, rect.top);
	int bottom = __min(rect.bottom + 1, image.GetHeight());
	for (int i = top; i < bottom; i++)
	{	for (int j = left; j < right; j++)
		{	BYTE* addr = (BYTE*)image.GetPixelAddress(j, i);
			*addr = GetRValue(color);
			*(addr + 1) = GetGValue(color);
			*(addr + 2) = GetBValue(color);
		}
	}
}

BOOL CImageMatrix::edgeDetect(CImageMatrix& im, 
double sigma, int* threshold, COLORREF color)
{
	if (image.IsNull())
		return FALSE;
	int width = image.GetWidth();
	int height = image.GetHeight();
	if (!im.createRGBImage(width, height))
		return FALSE;
	CEdgeDetector detector(getGrayBuffer(), width, height);
	BYTE* edgeMap = detector.edge_detect(sigma, threshold);
	if (edgeMap == 0)
		return FALSE;
	for (int i = 0; i < height; i++)
	{	for (int j = 0; j < width; j++)
		{	BYTE* addr = (BYTE*)im.image.GetPixelAddress(j, i);
			if (edgeMap[i * width + j] > 0)
			{	*addr = GetBValue(color);
				*(addr + 1) = GetGValue(color);
				*(addr + 2) = GetRValue(color);
			}
		}
	}
	return TRUE;
}

BOOL CImageMatrix::save(const char* fileName, REFGUID guidFileType)
{
	return SUCCEEDED(image.Save(fileName, guidFileType)) ? TRUE : FALSE;
}

BYTE* CImageMatrix::getGrayBuffer() const
{
	int width = image.GetWidth() - 1;
	int height = image.GetHeight() - 1;
	return getGrayBuffer(CRect(0, 0, width, height));
}

BYTE* CImageMatrix::getGrayBuffer(const CRect& region) const
{
	if (image.IsNull())
		return 0;
	int width = region.Width() + 1;
	int height = region.Height() + 1;
	BYTE* buffer = new BYTE[width * height];
	if (image.IsIndexed() && image.GetBPP() == BITS_PER_PIXEL_8)
	{	int maxColors = image.GetMaxColorTableEntries();
		RGBQUAD* rgbQuad = new RGBQUAD[maxColors];
		image.GetColorTable(0, maxColors, rgbQuad);
		for (int i = region.top; i <= region.bottom; i++)
		{	for (int j = region.left; j <= region.right; j++)
			{	BYTE addr = *((BYTE*)image.GetPixelAddress(j, i));
				BYTE red = rgbQuad[addr].rgbRed;
				BYTE green = rgbQuad[addr].rgbGreen;
				BYTE blue = rgbQuad[addr].rgbBlue;
				buffer[(i - region.top) * width + j - region.left] = 
					(red + green + blue) / 3;
			}
		}
		delete [] rgbQuad;
	}
	else if (image.GetBPP() == BITS_PER_PIXEL_24)
	{	for (int i = region.top; i <= region.bottom; i++)
		{	for (int j = region.left; j <= region.right; j++)
			{	BYTE* addr = (BYTE*)image.GetPixelAddress(j, i);
				BYTE red = *addr;
				BYTE green = *(addr + 1);
				BYTE blue = *(addr + 2);
				buffer[(i - region.top) * width + j - region.left] = 
					(red + green + blue) / 3;
			}
		}
	}
	return buffer;
}

BYTE* CImageMatrix::getColorBuffer(ColorType color) const
{
	if (image.IsNull())
		return 0;
	int width = image.GetWidth();
	int height = image.GetHeight();
	BYTE* buffer = new BYTE[width * height];
	for (int i = 0; i < height; i++)
	{	for (int j = 0; j < width; j++)
		{	BYTE* addr = (BYTE*)image.GetPixelAddress(j, i);
			if (color == Red)
				buffer[i * width + j] = *addr;
			else if (color == Green)
				buffer[i * width + j] = *(addr + 1);
			else if (color == Blue)
				buffer[i * width + j] = *(addr + 2);
		}
	}
	return buffer;
}

BOOL CImageMatrix::rotateImage(CImageMatrix& out, double angle)
{
	if (image.IsNull())
		return FALSE;
	if (image.IsIndexed() && image.GetBPP() == BITS_PER_PIXEL_8)
		rotate8BitImage(out, angle);
	else if (image.GetBPP() == BITS_PER_PIXEL_24)
		rotateRGBImage(out, angle);
	return TRUE;
}

void CImageMatrix::rotateRGBImage(CImageMatrix& out, double angle)
{
	long iX, iY, intX, intY;
	long widthIM1, heightIM1;
	double x, y, fractX, fractY;
	long x0 = image.GetWidth() / 2;
	long y0 = image.GetHeight() / 2;
	long heightI = (long)image.GetHeight();
	long widthI = (long)image.GetWidth();
	double cosA = cos(angle);
	double sinA = sin(angle);
	double startX = -(double)x0 * cosA + (double)y0 * sinA + (double)x0;
	double startY = -(double)x0 * sinA - (double)y0 * cosA + (double)y0;
	out.createRGBImage(widthI, heightI);
	widthIM1 = widthI - 1;
	heightIM1 = heightI - 1;
	BYTE *inAddr1, *inAddr2, *outAddr;
	for (iY = 0, y = startY; iY < heightI; iY++)
	{	for (iX = 0, x = startX; iX < widthI; iX++)
		{	intX = (long)x;
			intY = (long)y;
			outAddr = (BYTE*)out.image.GetPixelAddress(iX, iY);
			if ((intX >= 1 && intX < widthIM1) && (intY >= 1 && intY < heightIM1))
			{	fractX = x - (double)intX;
				fractY = y - (double)intY;
				if (fractX < 0.5)
				{	inAddr1 = (BYTE*)image.GetPixelAddress(intX, intY);
					inAddr2 = (BYTE*)image.GetPixelAddress(intX, intY + 1);
				}
				else
				{	inAddr1 = (BYTE*)image.GetPixelAddress(intX + 1, intY);
					inAddr2 = (BYTE*)image.GetPixelAddress(intX + 1, intY + 1);
				}
				if (fractY < 0.5)
				{	*outAddr = *inAddr1;
					*(outAddr + 1) = *(inAddr1 + 1);
					*(outAddr + 2) = *(inAddr1 + 2);
				}
				else
				{	*outAddr = *inAddr2;
					*(outAddr + 1) = *(inAddr2 + 1);
					*(outAddr + 2) = *(inAddr2 + 2);
				}
			}
			else
			{	*outAddr = 0;
				*(outAddr + 1) = 0;
				*(outAddr + 2) = 0;
			}
			x += cosA;
			y += sinA;
		}
		startY += cosA;
		y = startY;
		startX -= sinA;
	}
}

void CImageMatrix::rotate8BitImage(CImageMatrix& out, double angle)
{
	long iX, iY, intX, intY;
	long widthIM1, heightIM1;
	double x, y, fractX, fractY;
	long x0 = image.GetWidth() / 2;
	long y0 = image.GetHeight() / 2;
	long heightI = (long)image.GetHeight();
	long widthI = (long)image.GetWidth();
	double cosA = cos(angle);
	double sinA = sin(angle);
	double startX = -(double)x0 * cosA + (double)y0 * sinA + (double)x0;
	double startY = -(double)x0 * sinA - (double)y0 * cosA + (double)y0;
	out.createRGBImage(widthI, heightI);
	widthIM1 = widthI - 1;
	heightIM1 = heightI - 1;
	RGBQUAD* rgbQuad = 0;
	if (image.IsIndexed() && image.GetBPP() == BITS_PER_PIXEL_8)
	{	int maxColors = image.GetMaxColorTableEntries();
		rgbQuad = new RGBQUAD[maxColors];
		image.GetColorTable(0, maxColors, rgbQuad);
	}
	BYTE *inAddr1, *inAddr2, *outAddr;
	for (iY = 0, y = startY; iY < heightI; iY++)
	{	for (iX = 0, x = startX; iX < widthI; iX++)
		{	intX = (long)x;
			intY = (long)y;
			outAddr = (BYTE*)out.image.GetPixelAddress(iX, iY);
			if ((intX >= 1 && intX < widthIM1) && (intY >= 1 && intY < heightIM1))
			{	fractX = x - (double)intX;
				fractY = y - (double)intY;
				if (fractX < 0.5)
				{	inAddr1 = (BYTE*)image.GetPixelAddress(intX, intY);
					inAddr2 = (BYTE*)image.GetPixelAddress(intX, intY + 1);
				}
				else
				{	inAddr1 = (BYTE*)image.GetPixelAddress(intX + 1, intY);
					inAddr2 = (BYTE*)image.GetPixelAddress(intX + 1, intY + 1);
				}
				if (fractY < 0.5)
				{	*outAddr = rgbQuad[*inAddr1].rgbBlue;
					*(outAddr + 1) = rgbQuad[*inAddr1].rgbGreen;
					*(outAddr + 2) = rgbQuad[*inAddr1].rgbRed;
				}
				else
				{	*outAddr = rgbQuad[*inAddr2].rgbBlue;
					*(outAddr + 1) = rgbQuad[*inAddr2].rgbGreen;
					*(outAddr + 2) = rgbQuad[*inAddr2].rgbRed;
				}
			}
			else
			{	*outAddr = 0;
				*(outAddr + 1) = 0;
				*(outAddr + 2) = 0;
			}
			x += cosA;
			y += sinA;
		}
		startY += cosA;
		y = startY;
		startX -= sinA;
	}
	delete [] rgbQuad;
}

BOOL CImageMatrix::scaleImage(CImageMatrix& out, double xScale, double yScale)
{
	if (image.IsNull())
		return FALSE;
	if (image.IsIndexed() && image.GetBPP() == BITS_PER_PIXEL_8)
		scale8BitImage(out, xScale, yScale);
	else if (image.GetBPP() == BITS_PER_PIXEL_24)
		scaleRGBImage(out, xScale, yScale);
	return TRUE;
}

void CImageMatrix::scaleDim(double xScale, double yScale, int& rows, int& cols)
{
	double width = xScale * image.GetWidth();
	double height = yScale * image.GetHeight();
	double cWidth = ceil(width);
	double fWidth = floor(width);
	double cHeight = ceil(height);
	double fHeight = floor(height);
	if (fabs(cWidth - width) < SMALL)
		cols = (int)cWidth;
	else if (fabs(fWidth - width) < SMALL)
		cols = (int)fWidth;
	if (fabs(cHeight - height) < SMALL)
		rows = (int)cHeight;
	else if (fabs(fHeight - height) < SMALL)
		rows = (int)fHeight;
}

void CImageMatrix::scaleRGBImage(CImageMatrix& out, double xScale, double yScale)
{
	int i, j, k, x, y, rows, cols;
	int difX, difY, yLine, xZoom, yZoom;
	scaleDim(xScale, yScale, rows, cols);
	out.createRGBImage(cols, rows);
	xZoom = (int)(1.0 / xScale * 1000);
	yZoom = (int)(1.0 / yScale * 1000);
	y = 0;
	difY = 0;
	for (i = 0; i < image.GetHeight(); i++)
	{	difY -= 1000;
		x = 0;
		difX = 0;
		for (j = 0; j < image.GetWidth(); j++)
		{	difX -= 1000;
			while (difX < 0)
			{	if (x < cols)
				{	BYTE* inAddr = (BYTE*)image.GetPixelAddress(j, i);
					BYTE* outAddr = (BYTE*)out.image.GetPixelAddress(x, y);
					*outAddr = *inAddr;
					*(outAddr + 1) = *(inAddr + 1);
					*(outAddr + 2) = *(inAddr + 2);
				}
				x++;
				difX += xZoom;
			}
		}
		yLine = y;
		while (difY < 0)
		{	difY += yZoom;
			for (k = 0; k < x; k++)
			{	if (k < cols)
				{	BYTE* outAddrR = (BYTE*)out.image.GetPixelAddress(k, yLine);
					BYTE* outAddrL = (BYTE*)out.image.GetPixelAddress(k, y);
					*outAddrL = *outAddrR;
					*(outAddrL + 1) = *(outAddrR + 1);
					*(outAddrL + 2) = *(outAddrR + 2);
				}
			}
			y++;
			if (y >= rows)
				y = rows - 1;
		}
	}
}

void CImageMatrix::scale8BitImage(CImageMatrix& out, double xScale, double yScale)
{
	int i, j, k, x, y, rows, cols;
	int difX, difY, yLine, xZoom, yZoom;
	scaleDim(xScale, yScale, rows, cols);
	out.createRGBImage(cols, rows);
	xZoom = (int)(1.0 / xScale * 1000);
	yZoom = (int)(1.0 / yScale * 1000);
	y = 0;
	difY = 0;
	RGBQUAD* rgbQuad = 0;
	int maxColors = image.GetMaxColorTableEntries();
	rgbQuad = new RGBQUAD[maxColors];
	image.GetColorTable(0, maxColors, rgbQuad);
	for (i = 0; i < image.GetHeight(); i++)
	{	difY -= 1000;
		x = 0;
		difX = 0;
		for (j = 0; j < image.GetWidth(); j++)
		{	difX -= 1000;
			while (difX < 0)
			{	if (x < cols)
				{	BYTE* inAddr = (BYTE*)image.GetPixelAddress(j, i);
					BYTE* outAddr = (BYTE*)out.image.GetPixelAddress(x, y);
					*outAddr = rgbQuad[*inAddr].rgbBlue;
					*(outAddr + 1) = rgbQuad[*inAddr].rgbGreen;
					*(outAddr + 2) = rgbQuad[*inAddr].rgbRed;
				}
				x++;
				difX += xZoom;
			}
		}
		yLine = y;
		while (difY < 0)
		{	difY += yZoom;
			for (k = 0; k < x; k++)
			{	if (k < cols)
				{	BYTE* outAddrR = (BYTE*)out.image.GetPixelAddress(k, yLine);
					BYTE* outAddrL = (BYTE*)out.image.GetPixelAddress(k, y);
					*outAddrL = *outAddrR;
					*(outAddrL + 1) = *(outAddrR + 1);
					*(outAddrL + 2) = *(outAddrR + 2);
				}
			}
			y++;
			if (y >= rows)
				y = rows - 1;
		}
	}
	delete [] rgbQuad;
}

BOOL CImageMatrix::isNull() const
{
	return image.IsNull();
}

HBITMAP CImageMatrix::getHBitmap() const
{
	return image.operator HBITMAP();
}

int CImageMatrix::getWidth() const
{
	return !image.IsNull() ? image.GetWidth() : 0;
}

int CImageMatrix::getHeight() const
{
	return !image.IsNull() ? image.GetHeight() : 0;
}

BOOL CImageMatrix::demultiplex(const CImageMatrix& imL, 
const CImageMatrix& imR, CMatrix& out)
{
	if (imL.image.IsNull() || imR.image.IsNull())
		return FALSE;
	int rows = imL.getHeight();
	int cols = imL.getWidth();
	if (rows != imR.getHeight() || cols != imR.getWidth())
		return FALSE;
	out.SetSize(1, rows * cols * COLOR_CHANNEL_NUM * 2);
	for (int i = 0; i < rows; i++)
	{	for (int j = 0; j < cols; j++)
		{	BYTE* addrL = (BYTE*)imL.image.GetPixelAddress(j, i);
			BYTE* addrR = (BYTE*)imR.image.GetPixelAddress(j, i);
			BYTE lRed = *addrL;
			BYTE lGreen = *(addrL + 1);
			BYTE lBlue = *(addrL + 2);
			BYTE rRed = *addrR;
			BYTE rGreen = *(addrR + 1);
			BYTE rBlue = *(addrR + 2);
			out(0, 2 * i * cols + j) = lRed;
			out(0, 2 * i * cols + j + cols) = rRed;
			out(0, 2 * i * cols + j + 2 * cols * rows) = lGreen;
			out(0, 2 * i * cols + j + cols + 2 * cols * rows) = rGreen;
			out(0, 2 * i * cols + j + 2 * cols * 2 * rows) = lBlue;
			out(0, 2 * i * cols + j + cols + 2 * cols * 2 * rows) = rBlue;
		}
	}
	return TRUE;
}

BOOL CImageMatrix::multiplex(const CMatrix& m, 
CImageMatrix& imL, CImageMatrix& imR)
{
	int rows = (int)m.RowNo() / COLOR_CHANNEL_NUM;
	int cols = (int)m.ColNo() / 2;
	if (!imL.createRGBImage(cols, rows) || !imR.createRGBImage(cols, rows))
		return FALSE;
	for (int i = 0; i < rows; i++)
	{	for (int j = 0; j < cols; j++)
		{	BYTE* addrL = (BYTE*)imL.image.GetPixelAddress(j, i);
			BYTE* addrR = (BYTE*)imR.image.GetPixelAddress(j, i);
			*addrL = (BYTE)m(i, j);
			*(addrL + 1) = (BYTE)m(i + rows, j);
			*(addrL + 2) = (BYTE)m(i + 2 * rows, j);
			*addrR = (BYTE)m(i, j + cols);
			*(addrR + 1) = (BYTE)m(i + rows, j + cols);
			*(addrR + 2) = (BYTE)m(i + 2 * rows, j + cols);
		}
	}
	return TRUE;
}

void CImageMatrix::lowPassFilter(BYTE* imgIOn, int width, int height, 
int fltrSize, double cutOff)
{
	double stdDev;
	int i, midFltr, *fltr;
	stdDev = GAUSS_CUTOFF / cutOff;
	if (fltrSize == 0)
		fltrSize = (int)(3.0 / (2.0 * cutOff) + 0.5);
	if ((fltrSize % 2) == 0)
		fltrSize += 1;
	fltr = new int[fltrSize];
	midFltr = fltrSize / 2;
	for (i = 0; i <= midFltr; i++)
		fltr[i + midFltr] = fltr[midFltr - i]
			= (int)(NORM * exp((double)i * -i / (2.0 * stdDev * stdDev)) + 0.5);
	convolve(imgIOn, width, height, fltr, fltrSize);
}

void CImageMatrix::convolve(BYTE* imgIOn, int width, int height, 
int* fltr1D, int nFltr)
{
	long sumFltr, sum, midFltr, xEnd, yEnd, x, y, i;
	BYTE* imgInt = new BYTE[width * height];
	for (i = 0, sumFltr = 0; i < nFltr; i++)
		sumFltr += fltr1D[i];
	midFltr = (nFltr - 1) / 2;
	xEnd = width - midFltr;
	for (y = 0; y < height; y++)
	{	for (x = midFltr; x < xEnd; x++) 
		{	sum = fltr1D[midFltr] * imgIOn[y * width + x];
			for (i = 1; i <= midFltr; i++)
				sum += fltr1D[i + midFltr] * (imgIOn[y * 
					width + x - i] + imgIOn[y * width + x + i]);
			imgInt[y * width + x] = (BYTE)(sum / sumFltr);
		}
	}
	yEnd = height - midFltr;
	for (y = midFltr; y < yEnd; y++) 
	{	for (x = midFltr; x < xEnd; x++) 
		{	sum = fltr1D[midFltr] * imgInt[y * width + x];
			for (i = 1; i <= midFltr; i++)
				sum += fltr1D[i + midFltr] * (imgInt[(y - i) * 
					width + x] + imgInt[(y + i) * width + x]);
			imgIOn[y * width + x] = (BYTE)(sum / sumFltr);
		}
	}
	for (y = 0; y < height; y++)
		for (x = 0; x < midFltr; x++)
			imgIOn[y * width + x] = imgIOn[y * width + width - x - 1] = 0;
	for (y = 0; y < midFltr; y++)
		for (x = 0; x < width; x++)
			imgIOn[y * width + x] = imgIOn[(height - y - 1) * width + x] = 0;
	delete [] imgInt;
}