/***************************************************************************

PROJECT:		CS 298 Project

FILE:			winutil.cpp

PURPOSE:		Implementation that contains code to support the 
				display, querying, and traversal of directory and 
				file paths.

COMPILER:		Visual C++ .NET 2003

PROGRAMMER:		Wallun Chan

START DATE:		9/1/05

COMMENTS:		Code is taken from Mario Giannini and Jim Keogh. 
				Windows Programming: Programmer's Notebook. 
				Prentice Hall, 2001.

***************************************************************************/ 

#include "stdafx.h"
#include "winutil.h"
#include "constants.h"

/******* Functions to display, query, and traverse directories and files in MFC. *******/

int CStringCmp(const void* arg1, const void* arg2)
/* PURPOSE:		Comparison for two CStrings, for qsort call below

   RECEIVES:	Pointers to items to be compared.

   RETURNS:		-1 if arg1 < arg2, 1 if arg1 > arg2, and 0 if arg1 == arg2.

   REMARKS:		1) Assumes that function parameters have the proper
				   have the needed comparison operators.

				2) This code was borrowed from "Windows Programming:
				   Programmer's Notebook, by Mario Giannini and Jim Keogh,
				   2001; code found in PNDirTree.cpp in that book.
*/
{
	return (((CString*)arg1)->CompareNoCase(*(CString*)arg2));
}

void SlashPath(CString& Dest, bool MustHave)
/* PURPOSE:		Ensures that a path either has, or does not have,
				a slash at the end.

   RECEIVES:	CString object to slash or unslash depending on
				whether "MustHave" is true or false, respectively.

   REMARKS:		This code was borrowed from "Windows Programming:
				Programmer's Notebook, by Mario Giannini and Jim
				Keogh, 2001; code found in PNDirTree.cpp in that
				book.
*/
{
	char Now = '\0';
	if (!Dest.IsEmpty())
		Now = Dest[Dest.GetLength() - 1];
	if (MustHave && Now != SEPARATOR)
		Dest += SEPARATOR;
	else if (!MustHave && Now == SEPARATOR)
		Dest.TrimRight(SEPARATOR);
}

int BuildListEx(CStringArray& Dest, 
const char* Path, bool showFiles)
/* PURPOSE:		Builds a list of file and directory names for a path.

   RECEIVES:	CStringArray object to store file and directory names
				that are in "Path", and boolean "showFiles" to determine
				whether file names should also be stored.

   RETURNS:		Total number of files and directories in directory
				specified by "Path".

   REMARKS:		This code was borrowed and adapted from "Windows
				Programming: Programmer's Notebook, by Mario
				Giannini and Jim Keogh, 2001; code found in
				PNDirTree.cpp in that book.
*/
{
	CFileFind Find;
	CString MyPath = Path;
	BOOL Found;
	CWaitCursor Wait;
	SlashPath(MyPath, true);
	MyPath += "*.*";
	Found = Find.FindFile(MyPath);
	while (Found)
	{	Found = Find.FindNextFile();
		if (Find.IsDirectory() && !Find.IsDots())
			Dest.Add("A" + Find.GetFilePath());
		if (showFiles && !Find.IsDirectory())
			Dest.Add("B" + Find.GetFilePath());
	}
	qsort((void*)Dest.GetData(), Dest.GetSize(), sizeof(CString), CStringCmp);
	return (int)Dest.GetSize();
}

LPCSTR CWinUtil::GetSubPath(const char* Str)
/* PURPOSE:		To strip a path down to its file name.

   RECEIVES:	String name of full file path to process.

   RETURNS:		File name with path stripped.

   REMARKS:		This code was borrowed from "Windows Programming:
				Programmer's Notebook, by Mario Giannini and Jim
				Keogh, 2001; code found in PNDirTree.cpp in that
				book.
*/
{
	static CString Path;
	Path = Str;
	SlashPath(Path, false);
	int i = Path.ReverseFind(SEPARATOR);
	if (i != -1)
		Path = Path.Mid(i + 1);
	return Path;
}

int CWinUtil::BuildFileList(CStringArray& Dest, const char* Path)
/* PURPOSE:		Builds a list of file names for a path.

   RECEIVES:	CStringArray object to store file names that are in "Path".

   RETURNS:		Number of files in directory specified by "Path".

   REMARKS:		This code was borrowed and adapted from "Windows
				Programming: Programmer's Notebook, by Mario
				Giannini and Jim Keogh, 2001; code found in
				PNDirTree.cpp in that book.
*/
{
	int Ret = BuildListEx(Dest, Path, true);
	for (int i = 0; i < Ret; i++)
		Dest[i].Delete(0, 1);
	return (Ret);
}

int CWinUtil::BuildDirList(CStringArray& Dest, const char* Path)
/* PURPOSE:		Builds a list of file names for a path.

   RECEIVES:	CStringArray object to store directory names that are in "Path".

   RETURNS:		Number of directories in directory specified by "Path".

   REMARKS:		This code was borrowed and adapted from "Windows
				Programming: Programmer's Notebook, by Mario
				Giannini and Jim Keogh, 2001; code found in
				PNDirTree.cpp in that book.
*/
{
	int Ret = BuildListEx(Dest, Path, false);
	for (int i = 0; i < Ret; i++)
		Dest[i].Delete(0, 1);
	return (Ret);
}

BOOL CWinUtil::DemoBrowseForFolder(const char* Dest, 
const char* caption, bool showFiles)
/* PURPOSE:		Allows user to select a folder through a dialog GUI where
				directories and file names are displayed in a complete
				directory structure.

   RECEIVES:	Windows app handle hWnd, char string "Dest" to store
				user selected file or directory name, and boolean
				"showFiles" to determine whether files should be
				displayed in dialog GUI.

   RETURNS:		BOOL value determining whether user selected a file,
				or cancelled.

   REMARKS:		This code was borrowed and adapted from "Windows
				Programming: Programmer's Notebook, by Mario
				Giannini and Jim Keogh, 2001; code found in
				PNDirTree.cpp in that book.
*/
{
	BROWSEINFO bi;
	ITEMIDLIST* pItemIDList;
	IMalloc *pMalloc;
	if (CoGetMalloc(1, &pMalloc) != S_OK)
		return FALSE;
	char FolderOnly[_MAX_PATH];
	memset(&bi, 0, sizeof(bi));
	bi.hwndOwner = 0;
	bi.lpszTitle = caption;
	bi.pszDisplayName = FolderOnly;
	if (showFiles)
		bi.ulFlags = BIF_BROWSEINCLUDEFILES;
	if ((pItemIDList = SHBrowseForFolder(&bi)) != NULL)
	{	SHGetPathFromIDList(pItemIDList, const_cast<char*>(Dest));
		pMalloc->Free(pItemIDList);
		return TRUE;
	}
	return FALSE;
}